"""
Subpackage ``fits`` includes fits handling functions:
    - fits opening
    - fits info
    - fits writing
    - appending extensions to a fit file
    - ADI cube opening (cube with PA attached as HDU extension)
"""
from __future__ import absolute_import

from .fits import *
from .vipds9 import *

