"""
Subpackage ``exlib`` contains external libraries self contained in VIP:
IUWT from https://github.com/ratt-ru/PyMORESANE/
"""
from __future__ import absolute_import

#from .ds9 import *
from .iuwt import *

__all__ = []
